// CPUHOG.C - implements the CPUHOG functions.
// Author - Arnab Bhaduri
///////////////////////////////////////////////////////////////////

#define STRICT
#define NOGDI
#define NOIME

// Includes ///////////////////////////////////////////////////////

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <pdh.h>
#include "cpuhog.h"

// Local function prototypes //////////////////////////////////////

DWORD WINAPI   ThreadFunc( LPVOID lpv );

// Static variables ///////////////////////////////////////////////

static HANDLE ghThread     = NULL;   // worker thread handle
static HANDLE ghStopReqEvt = NULL;   // signals worker thread
                                     // to stop
static HANDLE ghStopAckEvt = NULL;   // worker thread acks the
                                     // stop request
static LONG   gCPUpc       = 50;     // CPU usage percentage
static int    gtSleep      = 1000;   // sleep interval in ms
static char   gszBuff[256];          // display msg buffer
static char   gszPdhKey[MAX_PATH];   // PDH query key

// SetHogProcessName - sets the process name used by the worker
// thread to monitor performance data. Can be set only when the
// worker thread is not running.
///////////////////////////////////////////////////////////////////
void SetHogProcessName( LPCSTR lpszName )
{
if( ghThread )
    return;
sprintf( gszPdhKey, "\\Process(%s)\\%% Processor Time", lpszName );
}

// GetHogStatusText - returns the current status text
///////////////////////////////////////////////////////////////////
LPCSTR GetHogStatusText( void )
{
return gszBuff;
}

// SetHogLevel - sets the CPU wastage percent level
///////////////////////////////////////////////////////////////////
void SetHogLevel( LONG lPercent )
{
gCPUpc = lPercent;
}

// StartHog - starts the CPU waster thread
///////////////////////////////////////////////////////////////////
BOOL StartHog( void )
{
DWORD dwID;

// start worker thread
ghThread = CreateThread( NULL, 0, ThreadFunc, NULL, 0, &dwID );
if( ghThread )
    {
    // set to highest possible priority
    SetPriorityClass( GetCurrentProcess(), HIGH_PRIORITY_CLASS );
    SetThreadPriority( ghThread, THREAD_PRIORITY_HIGHEST );
    }
else
    return FALSE;

// create the sync events to stop the worker thread
ghStopReqEvt = CreateEvent( NULL, FALSE, FALSE, NULL );
ghStopAckEvt = CreateEvent( NULL, FALSE, FALSE, NULL );

return TRUE;
}

// StopHog - stops the CPU waster thread
///////////////////////////////////////////////////////////////////
void StopHog( void )
{
if( ghStopReqEvt && ghStopAckEvt )
    {
    SetEvent( ghStopReqEvt );
    if( WaitForSingleObject(ghStopAckEvt, 15000) == WAIT_TIMEOUT )
        TerminateThread( ghThread, 0 );
    CloseHandle( ghStopReqEvt );
    CloseHandle( ghStopAckEvt );
    }
else
    TerminateThread( ghThread, 0 );

CloseHandle( ghThread );
ghThread = ghStopReqEvt = ghStopAckEvt = NULL;
}

// ThreadFunc - worker thread function, monitors and adjusts CPU
// utilization to the specified percentage.
///////////////////////////////////////////////////////////////////
DWORD WINAPI ThreadFunc( LPVOID lpv )
{
DWORD                dwDummy = 0;
HQUERY               hQuery  = NULL;
HCOUNTER             hCntr   = NULL;
PDH_RAW_COUNTER      first,
                     second;
PDH_FMT_COUNTERVALUE value;
PDH_STATUS           status;
long                 lDelta;
int                  i;

while( TRUE )
    {
    // check if we have been signalled to terminate
    if( WaitForSingleObject(ghStopReqEvt, 0) == WAIT_OBJECT_0 )
        {
        SetEvent( ghStopAckEvt );
        return 0;
        }

    // execute one query
    status = PdhOpenQuery( NULL, dwDummy, &hQuery );
    status = PdhAddCounter( hQuery, gszPdhKey, dwDummy, &hCntr );
    status = PdhCollectQueryData( hQuery );
    status = PdhGetRawCounterValue( hCntr, &dwDummy, &first );
    status = PdhCloseQuery( hQuery );

    // waste CPU time
    for( i = 0; i < 2; i++ )
        {
        Sleep( gtSleep / 4 );
        status = PdhOpenQuery( NULL, dwDummy, &hQuery );
        status = PdhAddCounter( hQuery, gszPdhKey, dwDummy, &hCntr );
        status = PdhCollectQueryData( hQuery );
        status = PdhGetRawCounterValue( hCntr, &dwDummy, &second );
        status = PdhCloseQuery( hQuery );
        }
    Sleep( gtSleep / 4 );

    // execute query again
    status = PdhOpenQuery( NULL, dwDummy, &hQuery );
    status = PdhAddCounter( hQuery, gszPdhKey, dwDummy, &hCntr );
    status = PdhCollectQueryData( hQuery );
    status = PdhGetRawCounterValue( hCntr, &dwDummy, &second );

    // compute the CPU utilization
    status =
        PdhCalculateCounterFromRawValue( hCntr, PDH_FMT_LONG,
                                         &second, &first, &value );
    status = PdhCloseQuery( hQuery );

    // check for success
    if( status != ERROR_SUCCESS )
        {
        sprintf( gszBuff, "Unable to query performance data." );
        return 1;
        }

    // update display text
    sprintf( gszBuff, 
             "Target CPU Usage: %ld %%\nCurrent CPU Usage %ld %%",
             gCPUpc,
             value.longValue );

    // adjust sleep time to reduce delta usage
    lDelta = abs(value.longValue - gCPUpc) * 40;
    if( lDelta <= 200 )
        lDelta = lDelta / 2;

    if( (value.longValue - gCPUpc) > 1 )
        gtSleep += lDelta;
    else if( (gCPUpc - value.longValue) > 1 )
        gtSleep -= lDelta;

    // complete sleep period
    Sleep( gtSleep / 4 );
    }

return 0;
}

//End //
